import time
import pprint
import inspect
import traceback
import threading
from typing import Callable
from functools import partial

import portalocker
from webview import window
from lightweight_charts import Chart

from lightweight_charts_server import ftype
from lightweight_charts_server.system import CallbackError
from lightweight_charts_server.system import init_render, log
from lightweight_charts_server.system import RENDER_JS, CHUNKS_DIR, CHUNKS_NUM


def inject_js(js_code: str):
    # This code is injected inside pywebview and its environment is not safe for concurrency.
    with portalocker.Lock(CHUNKS_NUM, timeout=1, mode="r+") as file:
        # Ensures chunks are created in the correct units and order
        before_chunk_num = int(file.read())
        chunk_num = before_chunk_num + 1
        chunk_filename = str(chunk_num) + ".js"
        # create chunk
        (CHUNKS_DIR / chunk_filename).write_text(js_code)
        # update index.js
        with RENDER_JS.open("a") as js:
            js.write("\n\n" + js_code)
        file.seek(0)
        file.write(str(chunk_num))
    # lightweight_charts sometimes generates invalid JS code,
    # which is likely due to the read/write operations being exposed to concurrency.
    # So, this also solves a fatal problem with lightweight_charts itself.


evaluate_js = window.Window.evaluate_js


def tracked_evaluate_js(self, js_code: str, *args, **kwargs):
    inject_js(js_code)
    return evaluate_js(self, js_code, *args, **kwargs)


window.Window.evaluate_js = tracked_evaluate_js


class View:

    def __init__(self, callback: Callable[..., Chart]):
        self.callback_origin = callback
        self.callback_signature = inspect.signature(callback)
        self.inspect_callback_signature()
        self.chart: Chart | None = None

    def inspect_callback_signature(self):
        """Validation of parameters defined in callback function signature"""

        for name, param in self.callback_signature.parameters.items():
            if param.annotation is inspect._empty:
                raise CallbackError(f"No type definition exists for parameter '{name}'")
            if not issubclass(param.annotation, ftype.FormType):
                raise CallbackError(
                    f"The type defined in the '{name}' parameter, "
                    f"{param.annotation}, is not supported"
                )
            if param.default is inspect._empty:
                raise CallbackError(f"No default value defined for parameter '{name}'")
            if not isinstance(param.default, param.annotation):
                raise CallbackError(
                    f"The type defined in the '{name}' parameter is different from the default type\n"
                    f"Detail: The type defined for the '{name}' parameter is {param.annotation}, "
                    f"but the type of the default value is {type(param.default)}."
                )

    def callback(self, params: dict) -> Chart:
        assert set(self.callback_signature.parameters.keys()) == set(params.keys())
        param_repr = (
            "---------- Parameters ----------\n\n"
            + pprint.pformat(params)
            + "\n\n--------------------------------"
        )
        try:
            start = time.time()
            result = self.callback_origin(**params)
            duration = time.time() - start
            log.info(
                f"Callback function executed in {duration:.2f} seconds\n" + param_repr
            )
        except Exception:
            raise CallbackError(
                "An error occurred in the callback function\n\n"
                f"{traceback.format_exc()}\n{param_repr}"
            )
        if not isinstance(result, Chart):
            raise CallbackError(
                "The callback function must return a Chart object.\n\n"
                f"Return of callback function: {result}"
            )
        return result

    def inject_form(self, params: dict):
        assert set(self.callback_signature.parameters.keys()) == set(params.keys())
        intput_tags = [
            f'<div class="input">{value.to_input(name)}</div>'
            for name, value in params.items()
        ]
        if intput_tags:
            inject_js(
                f""" 
                createCustomParameterSection(`
                    <form method="post" action="/parameter">
                        {"".join(intput_tags)}
                        <div class="submit">
                            <button type="submit">Apply</button>
                        </div>
                    </form>
                `)
            """
            )
        else:
            inject_js(
                f""" 
                createCustomParameterSection(`
                    <form>
                        <p>There are no parameters defined in the callback function.</p>
                    </form>
                `)
            """
            )

    def render(self, request: dict[str, str] = {}):
        if self.chart:
            self.chart.exit()
            del self.chart
        init_render()
        sig = self.callback_signature.parameters
        # A line of code to explicitly assign everything when running it for the first time.
        default = {name: param.default for name, param in sig.items()}
        update = {}

        for name, value in request.items():
            if sig[name].annotation in [ftype.DataFrame, ftype.JSON] and not value:
                # The default value of the file type is not passed to the Form.
                # Therefore, continue so that param.default can be used as is.
                continue
            update[name] = sig[name].annotation.from_input(value)

        params = default | update
        self.chart = self.callback(params)
        self.chart.show()
        self.inject_form(params)


class Stream:

    def __init__(
        self,
        chart: Chart,
        *,
        callback: Callable[[Chart], ...],
        latency: float = 0.1,
    ):
        assert isinstance(chart, Chart)
        self.chart = chart
        self.latency = latency
        self.callback_origin = callback
        self.callback_signature = inspect.signature(callback)
        self.inspect_callback_signature()

    def inspect_callback_signature(self):
        params = dict(self.callback_signature.parameters)
        if len(params) != 1:
            raise CallbackError(
                "The updater function must have one Chart type parameter."
            )
        if tuple(params.values())[0].annotation != Chart:
            raise CallbackError(
                "There is no Chart type annotation in the updater function parameter."
            )

    def callback(self):
        try:
            start = time.time()
            self.callback_origin(self.chart)
            duration = time.time() - start
            log.info(
                f"Callback function has finished. It ran for {duration:.2f} seconds."
            )
        except Exception:
            raise CallbackError(
                "An error occurred in the callback function\n\n"
                + traceback.format_exc()
            )

    def render(self):
        init_render()
        self.chart.show()
        self.thread = threading.Thread(target=partial(self.callback), daemon=True)
        self.thread.start()



import os
import shutil
import logging
from datetime import datetime
from pathlib import Path

import psutil

ROOT = Path(__file__).parent
STATIC_DIR = ROOT / "static"
RENDER_DIR = STATIC_DIR / "render"
RENDER_JS = RENDER_DIR / "index.js"
CHUNKS_DIR = RENDER_DIR / "chunks"
CHUNKS_NUM = RENDER_DIR / "chunk-num.lock"
LOG_TXT = RENDER_DIR / "log.txt"


def init_render():
    if RENDER_DIR.exists():
        for item in RENDER_DIR.iterdir():
            if item != LOG_TXT:
                if item.is_dir():
                    shutil.rmtree(item)
                else:
                    item.unlink()
    CHUNKS_DIR.mkdir(parents=True, exist_ok=True)
    RENDER_JS.write_text("")
    CHUNKS_NUM.write_text("0")


init_render()


class LogHandler(logging.NullHandler):
    """Expose to HTTP /static/render/log.txt"""

    pid = os.getpid()

    def __init__(self):
        super().__init__()

    def handle(self, record):
        now = datetime.now()
        memory = psutil.virtual_memory()
        memory_used = memory.total - memory.available
        memory_percent = (memory_used / memory.total) * 100

        disk = psutil.disk_usage("/")
        disk_used = disk.total - disk.free
        disk_percent = (disk_used / disk.total) * 100

        memory_percent = f"{memory_percent:.0f}%"
        disk_percent = f"{disk_percent:.0f}%"
        memory_gb = f"{memory_used * 1e-9:.0f}GB"
        disk_gb = f"{disk_used * 1e-9:.0f}GB"

        memory_status = f"Memory: {memory_gb}({memory_percent})"
        disk_status = f"Disk: {disk_gb}({disk_percent})"
        time = f"{now.month:02d}.{now.day:02d} {now.hour:02d}:{now.minute:02d}:{now.second:02d}"
        content = f"[{record.levelname}][{time}][pid:{self.pid}][{disk_status}][{memory_status}]: {self.format(record)}"
        print(content)
        with LOG_TXT.open("a") as log:
            log.write(content + "\n")


# Redirect output from the warnings module to the logging module.
logging.captureWarnings(True)
log = logging.getLogger("app")
log.propagate = False  # Prevent it from spreading to other logging outputs such as FastAPI or Uvicorn.
log.setLevel(logging.DEBUG)
formatter = logging.Formatter("%(message)s")
log_handler = LogHandler()
log_handler.setFormatter(formatter)
log_handler.setLevel(logging.DEBUG)
log.addHandler(log_handler)


class CallbackError(Exception):

    def __init__(self, msg=""):
        super().__init__(msg)
        self.msg = msg

    def __str__(self):
        content = (
            f"Invalid callback function\n\n"
            f"------------- Please fix the callback function as per the message below. -------------\n\n"
            f"{self.msg}"
            f"\n\n======================================================================================="
        )
        log.critical(content)
        return content



import asyncio

import uvicorn
from fastapi import FastAPI, Request, WebSocket
from fastapi.responses import HTMLResponse, FileResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from fastapi.websockets import WebSocketDisconnect

from lightweight_charts_server.display import View, Stream
from lightweight_charts_server.system import STATIC_DIR, CHUNKS_DIR, LOG_TXT


class Server:

    def __init__(
        self,
        display: View | Stream,
        host: str = "0.0.0.0",
        port: int = 80,
        title: str = "Chart",
        log_btn: bool = False,
    ):
        self.port = port
        self.host = host
        self.title = title
        self.log_btn = log_btn
        self.display = display
        self.display_type = display.__class__
        if self.display_type not in [View, Stream]:
            raise TypeError(f"{self.display_type} is not a valid display type.")

    async def root(self, request: Request):
        template = Jinja2Templates(directory=STATIC_DIR)
        return template.TemplateResponse(
            "index.html",
            {"request": request, "title": self.title, "log_btn": self.log_btn},
        )

    async def view_router(self, request: Request):
        self.display.render(await request.json())
        return {"result": "success"}

    async def stream_router(self, websocket: WebSocket):
        await websocket.accept()
        base_chunk_cnt = sum(1 for _ in CHUNKS_DIR.iterdir())
        while True:
            await asyncio.sleep(self.display.latency)
            chunks = sorted(
                [file for file in CHUNKS_DIR.iterdir()],
                key=lambda x: int(x.name.split(".")[0]),
            )
            chunk_cnt = len(chunks)
            if chunk_cnt > base_chunk_cnt:
                new_chunks = chunks[base_chunk_cnt - chunk_cnt :]
                script = "\n\n".join([chunk.read_text() for chunk in new_chunks])
                try:
                    await websocket.send_text(script)
                except WebSocketDisconnect:
                    break
                base_chunk_cnt = chunk_cnt

    async def get_log(self):
        return FileResponse(LOG_TXT)

    def serve(self):
        """This function should only be executed if __name__ == "__main__" """
        app = FastAPI()
        app.mount("/static", StaticFiles(directory=STATIC_DIR))
        app.get("/", response_class=HTMLResponse)(self.root)

        LOG_TXT.write_text("")
        app.get("/log")(self.get_log)

        try:
            if self.display_type == View:
                app.post("/view-parameter")(self.view_router)
                self.display.render()
            elif self.display_type == Stream:
                app.websocket("/stream")(self.stream_router)
                self.display.render()
        except RuntimeError as e:
            raise ImportError(
                "lightweight_charts_server must be imported before lightweight_charts"
            ) from e

        uvicorn.run(app, host=self.host, port=self.port)



import os
import sys
import platform
import subprocess
import multiprocessing

xvfb_error_message = """ 
Error: Xvfb is not installed or unavailable!

Please set up the Xvfb environment by installing the necessary packages using the following commands:

# Install Xvfb
apt install -y xvfb

# Install GTK (GUI Toolkit)
apt install -y python3-gi libgtk-3-dev gir1.2-webkit2-4.1

# Install the package that binds GTK to Python
apt install -y libgirepository1.0-dev 
pip install pygobject
"""

if platform.system() == "Linux" and not os.environ.get("DISPLAY"):
    try:
        subprocess.run(["xvfb-run", "-a", "python"] + sys.argv)
    except FileNotFoundError:
        print(xvfb_error_message)
    finally:
        sys.exit()

if multiprocessing.get_start_method() != "spawn":
    # If use Xvfb, All child processors must inherit the Xvfb context.
    multiprocessing.set_start_method("spawn", force=True)


from lightweight_charts_server.server import Server
from lightweight_charts_server.display import View, Stream
from lightweight_charts_server.system import log



"""
FormType: Classes that can define HTML Form creation
and HTTP request processing as Python function signatures
"""

import io
import json
from datetime import datetime
from abc import ABC, abstractmethod

import pandas as pd
from lightweight_charts_server.system import CallbackError


class FormType(ABC):

    @classmethod
    @abstractmethod
    def from_input(cls, value: str):
        """Creates an instance using the string received from the HTTP request."""

    @abstractmethod
    def to_input(self, name: str) -> str:
        """
        - Create an input HTML string to be placed inside the tag below.
        - `<div class="input">...</div>`
        """


class Int(int, FormType):
    """Inheritor of native int type"""

    @classmethod
    def from_input(cls, value: str):
        return cls(value)

    def to_input(self, name: str) -> str:
        return f"""
            <label for="{name}">{name.replace("_", " ")}</label>    
            <input name="{name}" type="number" value="{self}">
            """


class Float(float, FormType):
    """Inheritor of native float type"""

    @classmethod
    def from_input(cls, value: str):
        return cls(value)

    def to_input(self, name: str) -> str:
        return f"""
            <label for="{name}">{name.replace("_", " ")}</label>    
            <input name="{name}" type="number" value="{self}">
            """


class Str(str, FormType):
    """Inheritor of native str type"""

    @classmethod
    def from_input(cls, value: str):
        return cls(value)

    def to_input(self, name: str) -> str:
        return f"""
            <label for="{name}">{name.replace("_", " ")}</label>    
            <input name="{name}" type="text" value="{self}">
            """


class DateTime(datetime, FormType):
    """Inheritor of native datetime type"""

    @classmethod
    def from_input(cls, value: str):
        return cls.fromisoformat(value)

    def to_input(self, name: str) -> str:
        value = self.strftime("%Y-%m-%dT%H:%M:%S")
        return f"""
            <label for="{name}">{name.replace("_", " ")}</label>    
            <input name="{name}" type="datetime-local" value="{value}">
            """

    def __repr__(self):
        return f"<DateTime {self.isoformat()}>"


class DataFrame(pd.DataFrame, FormType):

    @classmethod
    def from_input(cls, value: str):
        if not value:
            return cls({})
        buffer = io.StringIO(value)
        df = pd.read_csv(buffer)
        return cls(df)

    def to_input(self, name: str) -> str:
        return f"""
            <label for="{name}">{name.replace("_", " ")}</label>
            <input type="file" name="{name}" accept=".csv">
            """

    def __repr__(self):
        rows, cols = self.shape
        memory_usage = self.memory_usage(deep=True).sum() / (1024**2)  # MB 단위로 변환
        summary = f"Rows: {rows}, Columns: {cols}, Memory Usage: {memory_usage:.3f} MB"
        return f"<DataFrame {summary}>"


class JSON(FormType):

    def __init__(self, obj: dict | list):
        self.obj = obj

    @classmethod
    def from_input(cls, value: str):
        if not value:
            return cls({})
        buffer = io.StringIO(value)
        obj = json.load(buffer)
        return cls(obj)

    def to_input(self, name: str) -> str:
        return f"""
            <label for="{name}">{name.replace("_", " ")}</label>
            <input type="file" name="{name}" accept=".json">
            """

    def __repr__(self):
        obj_str = str(self.obj)
        if len(obj_str) > 100:
            return f"<JSON {obj_str[:100]}...>"
        else:
            return f"<JSON {obj_str}>"


class Bool(FormType):

    def __init__(self, value: bool):
        assert isinstance(value, bool)
        self.value = value

    @classmethod
    def from_input(cls, value: str):
        return cls(bool(value))

    def to_input(self, name: str) -> str:
        return f"""
            <label for="{name}">{name.replace("_", " ")}</label>    
            <input name="{name}" type="checkbox" {"checked" if self.value else ""}>
            """

    def __repr__(self):
        return f"<Bool {self.value}>"


class Color(FormType):

    def __init__(self, hex_str: str):
        if not hex_str.startswith("#"):
            raise CallbackError(f"hex_str must start with '#'")
        hexadecimal = "0123456789abcdefABCDEF"
        _hex = hex_str[1:]
        if not len(_hex) in (3, 6) and all(s in hexadecimal for s in _hex):
            raise CallbackError(f"{hex_str} is not a valid hex color FormTypeat.")
        self.hex = hex_str

    @classmethod
    def from_input(cls, value: str):
        return cls(hex_str=value)

    def to_input(self, name: str):
        return f"""
            <label for="{name}">{name.replace("_", " ")}</label>    
            <input name="{name}" type="color" value="{self.hex}">
            """

    def __repr__(self):
        return f"<Color {self.hex}>"


class _Options(FormType):
    options: list[str]

    def __init__(self, selected: str):
        assert isinstance(selected, str)
        self.selected = selected
        if selected not in self.options:
            raise CallbackError(
                f"{selected}, an option not in {self.options}, is not allowed"
            )

    @classmethod
    def from_input(cls, value: str):
        return cls(value)

    def to_input(self, name: str):
        tags = []
        for option in self.options:
            if option == self.selected:
                tags.append(f'<option value="{option}" selected>{option}</option>')
            else:
                tags.append(f'<option value="{option}">{option}</option>')
        return f"""
            <label for="{name}">{name.replace("_", " ")}</label>
            <select name="{name}">{"".join(tags)}</select>
            """

    def __repr__(self):
        return f"<Options {self.options} selected='{self.selected}'>"


def options(*values: str) -> _Options:
    if all(not isinstance(value, str) for value in values):
        raise ValueError("options only accept strings")
    return type("Options", (_Options,), {"options": values})
